/* abc - The AspectBench Compiler
 * Copyright (C) 2006 Eric Bodden
 *
 * This compiler is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This compiler is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this compiler, in the file LESSER-GPL;
 * if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package abc.tm.weaving.weaver.tmanalysis;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import soot.jimple.toolkits.annotation.arraycheck.BoundedPriorityList;
import soot.toolkits.graph.Orderer;
import soot.toolkits.graph.SlowPseudoTopologicalOrderer;
import soot.toolkits.scalar.ForwardFlowAnalysis;
import abc.tm.weaving.aspectinfo.TraceMatch;
import abc.tm.weaving.matching.SMEdge;
import abc.tm.weaving.matching.TMStateMachine;
import abc.tm.weaving.weaver.tmanalysis.ShadowSMEdgeFactory.SMShadowEdge;
import abc.tm.weaving.weaver.tmanalysis.VariableSMEdgeFactory.SMVariableEdge;
import abc.tm.weaving.weaver.tmanalysis.ds.Configuration;
import abc.tm.weaving.weaver.tmanalysis.ds.ConfigurationBox;
import abc.tm.weaving.weaver.tmanalysis.ds.MemoryStableMRUCache;
import abc.tm.weaving.weaver.tmanalysis.query.Naming;

/**
 * This analysis computes a "may control flow" analysis for tracematches.
 * The bit that is important here to understand is the form of the directed graph that we analyze
 * and the analysis information we use. Opposed to standard flow analyses where we have an in-set and
 * an out-set for each <i>node</i> in a directed graph, we want to say something about <i>edges</i> here.
 * Hence, we model edges as nodes and nodes as edges. This is done by the adapter
 * {@link TMStateMachineAsGraph}, which does this conversion on-the-fly.
 * Our analysis information in the fixed point is for each edge the set of possible
 * tracematch transition that <i>could possibly be taken</i> at runtime when triggering the symbol
 * that is associated with this edge.<br>
 * Hence, the out-set of an edge p -s-> q holds all transitions p' -s-> q' for which
 * a transition n' -t-> p' is already in the in-set for some symbol t'.<br>
 * Unused edges can then be accessed via {@link #unusedEdgeIterator()}.
 * TODO additional comments about variables
 * @author Eric Bodden
 */
public class TraditionalTMMayFlowAnalysis extends ForwardFlowAnalysis implements TMMayFlowAnalysis {

	/**
	 * The state machine to interpret.
	 */
	protected final TMStateMachine stateMachine;	
	
	/**
	 * The program graph to interpret over.
	 */
	protected final TMStateMachine programGraph;

	protected final TraceMatch tracematch;	
	
	protected final Set activeShadows;
	
	protected final Configuration initialConfiguration;

	protected boolean analysisFinished;

	protected final boolean defaultOrder;
	
	protected Map edgeToBeforeFlowToAfterFlow = new IdentityHashMap();//new MemoryStableMRUCache("transitions",10*1024*1024,true);	

	protected Map leftToRightToMerge = new HashMap();//new MemoryStableMRUCache("merge",10*1024*1024,false); 

	/**
	 * Creates and performs a new flow analysis.
	 * @param tm a tracematch
	 * @param programGraph the program graph to iterate over 
	 * @param stateMachine the tracematch statemachine to use
	 * @param fullIteration if <code>true</code>, we do a full, unoptimized iteration
	 * @param defaultOrder if <code>true</code>, we use the (inappropriate) default iteration order
	 */
	public TraditionalTMMayFlowAnalysis(TraceMatch tm, TMStateMachine stateMachine, TMStateMachine programGraph, boolean fullIteration, boolean defaultOrder) {
		super(new TMStateMachineAsGraph(programGraph));
		this.defaultOrder = defaultOrder;

		//since we are iterating over state machine edges, which implement equals(..)
		//we need to separate those edges by using identity hash maps
		this.filterUnitToAfterFlow = new IdentityHashMap();
		this.filterUnitToBeforeFlow = new IdentityHashMap();
		this.unitToAfterFlow = new IdentityHashMap();
		this.unitToBeforeFlow = new IdentityHashMap();
		
		this.stateMachine = stateMachine;
		this.tracematch = tm;
		this.programGraph = programGraph;
		this.activeShadows = new HashSet();		

		//at this point really neither should be empty
		assert !this.stateMachine.isEmpty();
		assert !this.programGraph.isEmpty();
		
		//both graphs should be consistent
		assert this.stateMachine.isConsistent();
		assert this.programGraph.isConsistent();
		
		this.initialConfiguration = new Configuration(stateMachine.getStateIterator(),this);
		
		//do the analysis
		this.analysisFinished = false;
		doAnalysis();
		this.analysisFinished = true;
		
	}

	/**
	 * @param beforeFlow
	 * @param s
	 * @param afterFlow
	 */
	protected void flowThrough(Object beforeFlow, Object s, Object afterFlow) {
		Map beforeFlowToAfterFlow = (Map) edgeToBeforeFlowToAfterFlow.get(s);
		if(beforeFlowToAfterFlow==null) {
			beforeFlowToAfterFlow = new HashMap();
			edgeToBeforeFlowToAfterFlow.put(s,beforeFlowToAfterFlow);
		}
		ConfigurationBox cachedAfterFlow = (ConfigurationBox) beforeFlowToAfterFlow.get(beforeFlow);
		if(cachedAfterFlow==null) {
			cachedAfterFlow = new ConfigurationBox();
			doFlowThrough(beforeFlow, s, cachedAfterFlow);
			ConfigurationBox copyBeforeFlow = new ConfigurationBox();
			copy(beforeFlow,copyBeforeFlow);
			beforeFlowToAfterFlow.put(copyBeforeFlow, cachedAfterFlow);
//			Bench.FI9_TRANS_CACHE_MISSES++;
		}
//		else {
//			Bench.FI8_TRANS_CACHE_HITS++;
//		}
		copy(cachedAfterFlow,afterFlow);
	}

	
	/** 
	 * Computes the transformation of analysis information when flowing through an edge.
	 * The out-set of an edge p -s-> q holds all transitions p' -s-> q' for which
	 * a transition n' -t-> p' is in the in-set for some symbol t'.
	 * @param in the in-set
	 * @param d the current edge
	 * @param out the out-set
	 */
	protected void doFlowThrough(Object in, Object d, Object out) {
		ConfigurationBox cin = (ConfigurationBox) in;		
		SMEdge edge = (SMEdge) d;
		ConfigurationBox cout = (ConfigurationBox) out;		
		
		
		if(edge instanceof SMThreadSpawnEdge) {
			//if we have an edge indicating that a thread is spwaned
			SMThreadSpawnEdge threadSpawnEdge = (SMThreadSpawnEdge) edge;
			
			//if the thread may trigger any edges
			if(threadSpawnEdge.getThreadSummary().edgesMaybeTriggeredByThread().size()>0) {

				Configuration config = cin.get();
				
				Configuration newConfig = config.registerActiveThread(threadSpawnEdge);

				cout.set(newConfig);
				
			} else {
				//else just copy over
				copy(in, out);
			}
			
		} else if(edge.getLabel()==UGStateMachine.EPSILON ||
		   !Naming.getTracematchName(edge.getLabel()).equals(tracematch.getName())) {
			//if we have an epsilon transition or this is not the right tracematch, nothing changes
			copy(in, out);
		} else {		
			SMVariableEdge varEdge = (SMVariableEdge) edge;
			
			Configuration config = cin.get();

			Configuration newConfig = config.doTransition(varEdge);

			cout.set(newConfig);
		}
	}

	/**
	 * Registeres shadows as <i>active</i>, i.e. as being used in order
	 * to reach a final state.
	 * @param activeShadows a list of shadow IDs (int) for all active shadows
	 */
	public void registerActiveShadows(Set activeShadows) {
		if(analysisFinished) {
			throw new RuntimeException("This is only to be used from within the analysis.");			
		}
		this.activeShadows.addAll(activeShadows);
	}

	/**
	 * Returns the set of <i>active</i> shadows, i.e. shadows being used in order
	 * to reach a final state.
	 * @return a list of shadow IDs (int) for all active shadows
	 */
	public Set getActiveShadows() {
		return Collections.unmodifiableSet(activeShadows);
	}

	/**
	 * @return the tracematch
	 */
	public TraceMatch getTracematch() {
		return tracematch;
	}
	
	/** 
	 * {@inheritDoc}
	 */
	protected void copy(Object source, Object dest) {
		ConfigurationBox s = (ConfigurationBox) source;		
		ConfigurationBox d = (ConfigurationBox) dest;
		d.set(s.get());
	}

	/** 
	 * {@inheritDoc}
	 */
	protected Object entryInitialFlow() {
		ConfigurationBox initialFlow = new ConfigurationBox();
		initialFlow.set(initialConfiguration);
		return initialFlow;
	}

	/** 
	 * {@inheritDoc}
	 */
	protected void merge(Object in1, Object in2, Object out) {
		ConfigurationBox cin1 = (ConfigurationBox) in1;
		ConfigurationBox cin2 = (ConfigurationBox) in2;
		ConfigurationBox cout = (ConfigurationBox) out;
		
		if(cin1.isEmpty()) {
			copy(cin2,cout);
		} else if(cin2.isEmpty()) {
			copy(cin1,cout);
		} else {
			HashMap rightToMerge = (HashMap) leftToRightToMerge.get(in1);
			if(rightToMerge==null) {
				rightToMerge = new HashMap();
				ConfigurationBox copyin1 = new ConfigurationBox();
				copy(cin1,copyin1);
				leftToRightToMerge.put(copyin1, rightToMerge);
			}
			Object merge = rightToMerge.get(in2);
			if(merge==null) {
				Configuration c1 = cin1.get();
				Configuration c2 = cin2.get();
				assert c1.getStates().equals(c2.getStates());
				
				cout.set(c1.getJoinWith(c2));
				
				ConfigurationBox copyin2 = new ConfigurationBox();
				copy(cin2,copyin2);
				ConfigurationBox copyout = new ConfigurationBox();
				copy(cout,copyout);
				rightToMerge.put(copyin2, copyout);
//				Bench.FI7_MERGE_CACHE_MISSES++;
			} else {
				ConfigurationBox fmerge = (ConfigurationBox) merge;
				copy(fmerge,cout);
//				Bench.FI6_MERGE_CACHE_HITS++;
			}
		}
		
	}
	
	/** 
	 * {@inheritDoc}
	 */
	protected Object newInitialFlow() {
		return new ConfigurationBox();
	}

	/**
	 * Here we construct a special topological orderer which
	 * handles split points in a way that inner SCCs are computer first.
	 * @return a {@link SlowPseudoTopologicalOrderer}
	 */
	protected Orderer constructOrderer() {
		/* 
		 * Ordering seems to be important for the caching strategy.
		 */
		
		Orderer orderer;
		if(defaultOrder) {
			orderer = super.constructOrderer();
		} else {
			orderer = new SlowPseudoTopologicalOrderer();
		}
		//orderer = new FastOrderer();
		
		System.err.println("Using "+orderer.getClass().getName());
		return orderer;
	}
	
	/**
	 * Here we construct a worklist imlemented by a {@link BoundedPriorityList}.
	 * @return a {@link BoundedPriorityList}
	 */
	protected Collection constructWorklist(Map numbers) {
		if(defaultOrder) {
			return super.constructWorklist(numbers);
		}

		//we get a map object->index;
		//convert that into an ordered list
		List orderedList = new ArrayList(numbers.size());
		final Object SPACEHOLDER = new Object();
		for(int i=0;i<numbers.size();i++) {
			orderedList.add(SPACEHOLDER);
		}
		for (Iterator iter = numbers.entrySet().iterator(); iter.hasNext();) {
			Entry entry = (Entry) iter.next();
			Object oldVal = orderedList.set(((Integer)entry.getValue()).intValue()-1, entry.getKey());
			assert oldVal==SPACEHOLDER; //there should be no overrides
		}
		
		return new BoundedPriorityList(orderedList);
	}

	/**
	 * @return the stateMachine
	 */
	public TMStateMachine getStateMachine() {
		return stateMachine;
	}

}
