package abc.tm.weaving.weaver.tmanalysis;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import soot.Timers;
import soot.util.IdentityHashSet;
import abc.tm.weaving.aspectinfo.TraceMatch;
import abc.tm.weaving.matching.SMEdge;
import abc.tm.weaving.matching.SMNode;
import abc.tm.weaving.matching.TMStateMachine;
import abc.tm.weaving.weaver.tmanalysis.VariableSMEdgeFactory.SMVariableEdge;
import abc.tm.weaving.weaver.tmanalysis.ds.Configuration;
import abc.tm.weaving.weaver.tmanalysis.ds.Constraint;
import abc.tm.weaving.weaver.tmanalysis.ds.Disjunct;
import abc.tm.weaving.weaver.tmanalysis.query.Naming;
import abc.tm.weaving.weaver.tmanalysis.query.Shadow;
import abc.tm.weaving.weaver.tmanalysis.query.ShadowGroup;

/**
 * A customized, worklist-based algorithm which performs a {@link TMMayFlowAnalysis}.
 * It is usually faster and easier to understand than {@link TraditionalTMMayFlowAnalysis}. 
 * 
 * NOT thead safe due to static variables!
 *
 * @author Eric Bodden
 */
public class WorklistBasedTMFlowAnalysis implements TMMayFlowAnalysis {
	
	/** This is the principal data structure the algorithm operates on:
	 *  A mapping from states in the program graph to an abstract tracematch configuration
	 *  at this state. Those configurations will evolve over time until a fixed point is reached. */
	protected IdentityHashMap stateToConfig;
	
	/** the set of active shadows discovered by the analysis, i.e. the result set */
	protected Set activeShadows;

	/** the set of states of the program graph for which a predecessor has changed
	 * and for which the configuration has to be recomputed */
	protected Set worklist;

	/**
	 * a most-recently-used cache for caching merge resuts; it stores results of the merge
	 * function {@link Configuration} -> {@link Configuration} -> {@link Configuration} */
	protected Map leftToRightToMerge = new HashMap();//new MemoryStableMRUCache("merge",10*1024*1024,false);
	
	/** 
	 *  a most-recently used cache to cache the result of the transition function, which has type
	 *  {@link Configuration} -> {@link SMEdge} -> {@link Configuration}; the cache stores a mapping
	 *  {@link SMEdge} -> {@link Configuration} -> {@link Configuration} for improved performance (edges are unique while
	 *  configurations are not necessarily)
	 * */
	protected Map edgeToConfigToSuccConfig = new IdentityHashMap();// new MemoryStableMRUCache("transitions",10*1024*1024,true);

	/** the associated tracematch */
	protected final TraceMatch tm;
	
	/** shortcut to the associated state machine */
	protected final TMStateMachine sm;

	/** handle to the program graph */
	protected final TMStateMachine programGraph; 

	/** internal flag to check that no shadows are registered after the analysis finished */
	protected boolean analysisFinished;

	protected final ShadowGroup group;
	
	protected final Collection shadowScope;
	
	protected final static int MAX_NUM_DISJUNCTS = 30; 
	
	/**
	 * Creates a new flow-sensitive analysis for the given tracematch and a program graph.
	 * @param group a {@link ShadowGroup}
	 * @param programGraph a flow-sensitive representation of the program to analyze in form of a finite state machine
	 * @param onlyConsiderLabelShadows if true, the iteration only takes label shadows and no skip shadows into account
	 * @see UGStateMachine#fold()
	 */
	public WorklistBasedTMFlowAnalysis(ShadowGroup group, TMStateMachine programGraph, boolean onlyConsiderLabelShadows) {		
		this.group = group;
		this.tm = group.getTraceMatch();
		this.sm = (TMStateMachine) this.tm.getStateMachine();
		this.activeShadows = new HashSet();
		this.analysisFinished = false;
		this.programGraph = programGraph;
		
		if(onlyConsiderLabelShadows) {
			//if this flag is set, automatically retain all skip shadows but reduce the scope to only label shadows
			this.activeShadows.addAll(Shadow.uniqueShadowIDsOf(group.getSkipShadows()));
			this.shadowScope = group.getLabelShadows(); 
		} else {
			//else iterate over all shadows
			this.shadowScope = group.getAllShadows(); 
		}
		
		doAnalysis();
			
		reset();
		this.analysisFinished = true;
	}

	/**
	 * 
	 */
	private static void reset() {
		Configuration.reset();
		Constraint.reset();
		Disjunct.reset();
	}

	/**
	 * Executes the analysis.
	 */
	protected void doAnalysis() {
		stateToConfig = new IdentityHashMap();
		worklist = new IdentityHashSet();
		
		//create the initial configuration
		Configuration initial = new Configuration(sm.getStateIterator(),this);
		
		//push all initial states of the program graph (usually just one) on the worklist and
		//assiciate an initial configuration with them
		for (Iterator stateIter = programGraph.getStateIterator(); stateIter.hasNext();) {
			SMNode state = (SMNode) stateIter.next();
			if(state.isInitialNode()) {				
				stateToConfig.put(state, initial);
				push(state);
			}
		}
		
		//while there are states left
		while(!worklist.isEmpty()) {
			SMNode state = pop();
			
			//get the associated configuration
			Configuration config = (Configuration) stateToConfig.get(state); 

			//for all edges out of the state...
			for (Iterator outIter = state.getOutEdgeIterator(); outIter.hasNext();) {
				SMEdge outEdge = (SMEdge) outIter.next();
				
				//calculate the transition; as a side-effect, when the configuration is driven into
				//a final state, the corresponding disjunct's history will be comitted using
				//registerActiveShadows(Set)
				Configuration newConfig = transition(config,outEdge);

				//merge the resulting configuration into the configuration of target state
				SMNode to = outEdge.getTarget();
				mergeTo(to,newConfig);
			}
			
		}
		
		//done; we don't need those any more
		stateToConfig = null;
		leftToRightToMerge = null;
		edgeToConfigToSuccConfig = null;
		worklist = null;
		stateToConfig = null;
	}

	/**
	 * Calculates the successor configuration of config when taking the edge.
	 * The result is cached internally.
	 * @param config any {@link Configuration}
	 * @param edge any {@link SMEdge}
	 * @return the resulting successor configuration
	 */
	protected Configuration transition(Configuration config, SMEdge edge) {

		Map configToSuccessor = (Map) edgeToConfigToSuccConfig.get(edge);
		if(configToSuccessor==null) {
			configToSuccessor = new HashMap();
			edgeToConfigToSuccConfig.put(edge,configToSuccessor);
		}
		Configuration cachedResult = (Configuration) configToSuccessor.get(config);
		if(cachedResult==null) {
			cachedResult = doTransition(config,edge);
			configToSuccessor.put(config, cachedResult);
		} 
		return cachedResult;
	}
	
	/**
	 * Calculates the successor configuration of config when taking the edge.
	 * @param config any {@link Configuration}
	 * @param edge any {@link SMEdge}
	 * @return the resulting successor configuration
	 */
	protected Configuration doTransition(Configuration config, SMEdge edge) {
		
		//for statistics
		Timers.v().totalFlowComputations++;
		
		if(edge instanceof SMThreadSpawnEdge) {
			//if we have an edge indicating that a thread is spwaned
			SMThreadSpawnEdge threadSpawnEdge = (SMThreadSpawnEdge) edge;
			
			throw new IllegalStateException("not supported any more");
			
//			//if the thread may trigger any edges
//			if(threadSpawnEdge.getThreadSummary().edgesMaybeTriggeredByThread().size()>0) {
//
//				//register any active threads
//				Configuration newConfig = config.registerActiveThread(threadSpawnEdge);
//				return newConfig;				
//				
//			} else {
//				//else just copy over
//				return config;
//			}
			
		} else if(edge.getLabel()==UGStateMachine.EPSILON ||
		   !Naming.getTracematchName(edge.getLabel()).equals(tm.getName())) {
			//if we have an epsilon transition or this is not the right tracematch, nothing changes
			return config;
		} else {		
			//if we have a shadow edge
			SMVariableEdge varEdge = (SMVariableEdge) edge;
			
			assert varEdge.hasShadow();
			
			if(shadowScope.contains(varEdge.getShadow())) {
				//compute the transition normally
				Configuration newConfig = config.doTransition(varEdge);
				return newConfig;			
			} else {
				return config;
			}
			
		}
		
	}

	/**
	 * Merges the configuration of the state with the one that is passed in.
	 * @param state any {@link SMNode} of the {@link #programGraph}
	 * @param newConfig a {@link Configuration} of a predecessor node of state
	 */
	protected void mergeTo(SMNode state, Configuration newConfig) {
		Configuration oldConfig = (Configuration) stateToConfig.get(state);
		if(oldConfig==null) {
			//nothing to merge, just copy
			stateToConfig.put(state, newConfig);			
			push(state);
		} else {
			//actually merge
			Configuration merged = merge(newConfig, oldConfig);
			stateToConfig.put(state, merged);
			if(!oldConfig.equals(merged)) {
				push(state);
			}
		}
	}

	/**
	 * Merges the two configurations, returning the merge result.
	 * The result is cached.
	 * @param left any {@link Configuration} that has the same state set as right
	 * @param right any {@link Configuration} that has the same state set as left
	 * @return the result of a merge of left and right
	 */
	protected Configuration merge(Configuration left, Configuration right) {
		HashMap rightToMerge = (HashMap) leftToRightToMerge.get(left);
		if(rightToMerge==null) {
			rightToMerge = new HashMap();
			leftToRightToMerge.put(left, rightToMerge);
		}
		Object merge = rightToMerge.get(right);
		if(merge==null) {
			assert left.getStates().equals(right.getStates());		
			Configuration merged = left.getJoinWith(right);			
			rightToMerge.put(right, merged);
			return merged;
		} else {
			return (Configuration) merge;			
		}
	}

	/**
	 * Pushes the state on the {@link #worklist}.
	 * @param state any {@link SMNode} of the {@link #programGraph}
	 */
	private void push(SMNode state) {
		worklist.add(state);
	}

	/**
	 * Pops a state from the {@link #worklist}. Clients have to make sure that
	 * the {@link #worklist} is not empty.
	 * @return an element from the {@link #worklist} 
	 */
	private SMNode pop() {
		Iterator iterator = worklist.iterator();
		SMNode next = (SMNode) iterator.next();
		iterator.remove();
		return next;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public TraceMatch getTracematch() {
		return tm;
	}

	/**
	 * {@inheritDoc}
	 */
	public void registerActiveShadows(Set activeShadows) {
		if(analysisFinished) {
			throw new RuntimeException("This is only to be used from within the analysis.");			
		}
		this.activeShadows.addAll(activeShadows);
	}

	/**
	 * {@inheritDoc}
	 */
	public Set getActiveShadows() {
		return Collections.unmodifiableSet(activeShadows);
	}

}
